#include <vector>
#include <set>

#include "Pcl6BaseVisitor.h"
#include "antlr4-runtime.h"

#include "intermediate/symtab/SymtabStack.h"
#include "intermediate/symtab/Symtab.h"
#include "intermediate/symtab/SymtabEntry.h"
#include "intermediate/symtab/Predefined.h"
#include "intermediate/type/Typespec.h"
#include "intermediate/type/TypeChecker.h"
#include "intermediate/util/CrossReferencer.h"
#include "SemanticErrorHandler.h"
#include "Semantics.h"

namespace intermediate { namespace symtab {
    int Symtab::unnamedIndex = 0;
}}

namespace frontend {

using namespace std;
using namespace intermediate::symtab;
using namespace intermediate::type;
using namespace intermediate::util;

Object Semantics::visitProgram(Pcl6Parser::ProgramContext *ctx)
{
    visit(ctx->programHeader());
    visit(ctx->block()->declarations());
    visit(ctx->block()->compoundStatement());

    // Print the cross-reference table.
    CrossReferencer crossReferencer;
    crossReferencer.print(symtabStack);

    return nullptr;
}

Object Semantics::visitProgramHeader(Pcl6Parser::ProgramHeaderContext *ctx)
{
    Pcl6Parser::ProgramIdentifierContext *idCtx = ctx->programIdentifier();
    string programName = idCtx->IDENTIFIER()->getText();  // don't shift case

    programId = symtabStack->enterLocal(programName, PROGRAM);
    programId->setRoutineSymtab(symtabStack->push());

    symtabStack->setProgramId(programId);
    symtabStack->getLocalSymtab()->setOwner(programId);

    idCtx->entry = programId;
    return nullptr;
}

Object Semantics::visitConstantDefinition(
                                Pcl6Parser::ConstantDefinitionContext *ctx)
{
    Pcl6Parser::ConstantIdentifierContext *idCtx = ctx->constantIdentifier();
    string constantName = toLowerCase(idCtx->IDENTIFIER()->getText());
    SymtabEntry *constantId = symtabStack->lookupLocal(constantName);

    if (constantId == nullptr)
    {
        Pcl6Parser::ConstantContext *constCtx = ctx->constant();
        Object constValue = visit(constCtx);

        constantId = symtabStack->enterLocal(constantName, CONSTANT);
        constantId->setValue(constValue);
        constantId->setType(constCtx->type);

        idCtx->entry = constantId;
        idCtx->type  = constCtx->type;
    }
    else
    {
        error.flag(REDECLARED_IDENTIFIER, ctx);

        idCtx->entry = constantId;
        idCtx->type  = Predefined::integerType;
    }

    constantId->appendLineNumber(ctx->getStart()->getLine());
    return nullptr;
}

Object Semantics::visitConstant(Pcl6Parser::ConstantContext *ctx)
{
    if (ctx->IDENTIFIER() != nullptr)
    {
        string constantName = toLowerCase(ctx->IDENTIFIER()->getText());
        SymtabEntry *constantId = symtabStack->lookup(constantName);

        if (constantId != nullptr)
        {
            Kind kind = constantId->getKind();
            if ((kind != CONSTANT) && (kind != ENUMERATION_CONSTANT))
            {
                error.flag(INVALID_CONSTANT, ctx);
            }

            ctx->type  = constantId->getType();
            ctx->value = constantId->getValue();

            constantId->appendLineNumber(ctx->getStart()->getLine());
        }
        else
        {
            error.flag(UNDECLARED_IDENTIFIER, ctx);
            ctx->type = Predefined::integerType;
            ctx->value = 0;
        }
    }
    else if (ctx->characterConstant() != nullptr)
    {
        ctx->type = Predefined::charType;
        ctx->value = ctx->getText()[1];
    }
    else if (ctx->stringConstant() != nullptr)
    {
        string pascalString = ctx->stringConstant()->STRING()->getText();
        string unquoted = pascalString.substr(1, pascalString.length()-2);

        size_t pos = 0;
        while ((pos = unquoted.find("''", pos)) != string::npos)
        {
            unquoted.replace(pos++, 2, "'");
        }

        pos = 0;
        while ((pos = unquoted.find("\"", pos)) != string::npos)
        {
            unquoted.replace(pos++, 2, "\\\"");
        }

        ctx->type = Predefined::stringType;
        ctx->value = new string(unquoted);
    }
    else  // number
    {
        Pcl6Parser::UnsignedNumberContext *unsignedCtx = ctx->unsignedNumber();

        if (unsignedCtx->integerConstant() != nullptr)
        {
            ctx->type = Predefined::integerType;
            ctx->value = stoi(ctx->getText());
        }
        else
        {
            ctx->type = Predefined::realType;
            ctx->value = stod(ctx->getText());
        }
    }

    return Object(ctx->value);
}

Object Semantics::visitTypeDefinition(Pcl6Parser::TypeDefinitionContext *ctx)
{
    Pcl6Parser::TypeIdentifierContext *idCtx = ctx->typeIdentifier();
    string typeName = toLowerCase(idCtx->IDENTIFIER()->getText());
    SymtabEntry *typeId = symtabStack->lookupLocal(typeName);

    Pcl6Parser::TypeSpecificationContext *typespecCtx =
                                                    ctx->typeSpecification();

    // If it's a record type, create a named record type.
    Pcl6Parser::RecordTypespecContext *recordTypespecCtx =
                dynamic_cast<Pcl6Parser::RecordTypespecContext *>(typespecCtx);
    if (recordTypespecCtx != nullptr)
    {
        typeId = createRecordType(recordTypespecCtx, typeName);
    }

    // Enter the type name of any other type into the symbol table.
    else if (typeId == nullptr)
    {
        Pcl6Parser::TypeSpecificationContext *typeCtx =
                                                    ctx->typeSpecification();
        visit(typeCtx);

        typeId = symtabStack->enterLocal(typeName, TYPE);
        typeId->setType(typeCtx->type);
        typeCtx->type->setIdentifier(typeId);
    }
    else
    {
        error.flag(REDECLARED_IDENTIFIER, ctx);
    }

    idCtx->entry = typeId;
    idCtx->type  = typespecCtx->type;

    typeId->appendLineNumber(ctx->getStart()->getLine());
    return nullptr;
}

Object Semantics::visitRecordTypespec(Pcl6Parser::RecordTypespecContext *ctx)
{
    // Create an unnamed record type.
    string recordTypeName = Symtab::generateUnnamedName();
    createRecordType(ctx, recordTypeName);

    return nullptr;
}

SymtabEntry *Semantics::createRecordType(
                    Pcl6Parser::RecordTypespecContext *recordTypeSpecCtx,
                    string recordTypeName)
{
    Pcl6Parser::RecordTypeContext *recordTypeCtx =
                                            recordTypeSpecCtx->recordType();
    Typespec *recordType = new Typespec(RECORD);

    SymtabEntry *recordTypeId = symtabStack->enterLocal(recordTypeName, TYPE);
    recordTypeId->setType(recordType);
    recordType->setIdentifier(recordTypeId);

    string recordTypePath = createRecordTypePath(recordType);
    recordType->setRecordTypePath(recordTypePath);

    // Enter the record fields into the record type's symbol table.
    Symtab *recordSymtab = createRecordSymtab(recordTypeCtx->recordFields(),
                                             recordTypeId);
    recordType->setRecordSymtab(recordSymtab);

    recordTypeCtx->entry    = recordTypeId;
    recordTypeSpecCtx->type = recordType;

    return recordTypeId;
}

string Semantics::createRecordTypePath(Typespec *recordType)
{
    SymtabEntry *recordId = recordType->getIdentifier();
    SymtabEntry *parentId = recordId->getSymtab()->getOwner();
    string path = recordId->getName();

    while (   (parentId->getKind() == TYPE)
           && (parentId->getType()->getForm() == RECORD))
    {
        path = parentId->getName() + "$" + path;
        parentId = parentId->getSymtab()->getOwner();
    }

    path = parentId->getName() + "$" + path;
    return path;
}

Symtab *Semantics::createRecordSymtab(
                Pcl6Parser::RecordFieldsContext *ctx, SymtabEntry *ownerId)
{
    Symtab *recordSymtab = symtabStack->push();

    recordSymtab->setOwner(ownerId);
    visit(ctx->variableDeclarationsList());
    recordSymtab->resetVariables(RECORD_FIELD);
    symtabStack->pop();

    return recordSymtab;
}

Object Semantics::visitSimpleTypespec(Pcl6Parser::SimpleTypespecContext *ctx)
{
    visit(ctx->simpleType());
    ctx->type = ctx->simpleType()->type;

    return nullptr;
}

Object Semantics::visitTypeIdentifierTypespec(
                            Pcl6Parser::TypeIdentifierTypespecContext *ctx)
{
    visit(ctx->typeIdentifier());
    ctx->type = ctx->typeIdentifier()->type;

    return nullptr;
}

Object Semantics::visitTypeIdentifier(Pcl6Parser::TypeIdentifierContext *ctx)
{
    string typeName = toLowerCase(ctx->IDENTIFIER()->getText());
    SymtabEntry *typeId = symtabStack->lookup(typeName);

    if (typeId != nullptr)
    {
        if (typeId->getKind() != TYPE)
        {
            error.flag(INVALID_TYPE, ctx);
            ctx->type = Predefined::integerType;
        }
        else
        {
            ctx->type = typeId->getType();
        }

        typeId->appendLineNumber(ctx->start->getLine());
    }
    else
    {
        error.flag(UNDECLARED_IDENTIFIER, ctx);
        ctx->type = Predefined::integerType;
    }

    ctx->entry = typeId;
    return nullptr;
}

Object Semantics::visitEnumerationTypespec(
                                Pcl6Parser::EnumerationTypespecContext *ctx)
{
    Typespec *enumType = new Typespec(ENUMERATION);
    vector<SymtabEntry *> *constants = new vector<SymtabEntry *>();
    int value = -1;

    // Loop over the enumeration constants.
    for (Pcl6Parser::EnumerationConstantContext *constCtx :
                                ctx->enumerationType()->enumerationConstant())
    {
        Pcl6Parser::ConstantIdentifierContext *constIdCtx =
                                                constCtx->constantIdentifier();
        string constantName = toLowerCase(constIdCtx->IDENTIFIER()->getText());
        SymtabEntry *constantId = symtabStack->lookupLocal(constantName);

        if (constantId == nullptr)
        {
            constantId = symtabStack->enterLocal(constantName,
                                                 ENUMERATION_CONSTANT);
            constantId->setType(enumType);
            constantId->setValue(++value);

            constants->push_back(constantId);
        }
        else
        {
            error.flag(REDECLARED_IDENTIFIER, constCtx);
        }

        constIdCtx->entry = constantId;
        constIdCtx->type  = enumType;

        constantId->appendLineNumber(ctx->getStart()->getLine());
    }

    enumType->setEnumerationConstants(constants);
    ctx->type = enumType;

    return nullptr;
}

Object Semantics::visitSubrangeTypespec(
                                    Pcl6Parser::SubrangeTypespecContext *ctx)
{
    Typespec *type = new Typespec(SUBRANGE);
    Pcl6Parser::SubrangeTypeContext *subCtx = ctx->subrangeType();
    Pcl6Parser::ConstantContext *minCtx = subCtx->constant()[0];
    Pcl6Parser::ConstantContext *maxCtx = subCtx->constant()[1];

    Object minObj = visit(minCtx);
    Object maxObj = visit(maxCtx);

    Typespec *minType = minCtx->type;
    Typespec *maxType = maxCtx->type;

    if (   (   (minType->getForm() != SCALAR)
    		&& (minType->getForm() != ENUMERATION))
        || (minType == Predefined::realType)
        || (minType == Predefined::stringType))
    {
        error.flag(INVALID_CONSTANT, minCtx);
        minType = Predefined::integerType;
        minObj  = 0;
    }

    int minValue;
    int maxValue;

    if (minType == Predefined::integerType)
    {
    	minValue = minObj.as<int>();
    	maxValue = maxObj.as<int>();
    }
    else if (minType == Predefined::charType)
    {
    	minValue = minObj.as<char>();
    	maxValue = maxObj.as<char>();
    }
    else  // enumeration constants
    {
    	minValue = minCtx->value.as<int>();
    	maxValue = maxCtx->value.as<int>();
    }

    if ((maxType != minType) || (minValue > maxValue))
    {
        error.flag(INVALID_CONSTANT, maxCtx);
        maxType = minType;
        maxObj  = minObj;
    }

    type->setSubrangeBaseType(minType);
    type->setSubrangeMinValue(minValue);
    type->setSubrangeMaxValue(maxValue);

    ctx->type = type;
    return nullptr;
}

Object Semantics::visitArrayTypespec(Pcl6Parser::ArrayTypespecContext *ctx)
{
    Typespec *arrayType = new Typespec(ARRAY);
    Pcl6Parser::ArrayTypeContext *arrayCtx = ctx->arrayType();
    Pcl6Parser::ArrayDimensionListContext *listCtx =
                                                arrayCtx->arrayDimensionList();

    ctx->type = arrayType;

    // Loop over the array dimensions.
    int count = listCtx->simpleType().size();
    for (int i = 0; i < count; i++)
    {
        Pcl6Parser::SimpleTypeContext *simpleCtx = listCtx->simpleType()[i];
        visit(simpleCtx);
        arrayType->setArrayIndexType(simpleCtx->type);
        arrayType->setArrayElementCount(typeCount(simpleCtx->type));

        if (i < count-1)
        {
            Typespec *elmtType = new Typespec(ARRAY);
            arrayType->setArrayElementType(elmtType);
            arrayType = elmtType;
        }
    }

    visit(arrayCtx->typeSpecification());
    Typespec *elmtType = arrayCtx->typeSpecification()->type;
    arrayType->setArrayElementType(elmtType);

    return nullptr;
}

int Semantics::typeCount(Typespec *type)
{
    int count = 0;

    if (type->getForm() == ENUMERATION)
    {
        vector<SymtabEntry *> *constants = type->getEnumerationConstants();
        count = constants->size();
    }
    else  // subrange
    {
        int minValue = type->getSubrangeMinValue();
        int maxValue = type->getSubrangeMaxValue();
        count = maxValue - minValue + 1;
    }

    return count;
}

Object Semantics::visitVariableDeclarations(
                                Pcl6Parser::VariableDeclarationsContext *ctx)
{
    Pcl6Parser::TypeSpecificationContext *typeCtx = ctx->typeSpecification();
    visit(typeCtx);

    Pcl6Parser::VariableIdentifierListContext *listCtx =
                                                ctx->variableIdentifierList();

    // Loop over the variables being declared.
    for (Pcl6Parser::VariableIdentifierContext *idCtx :
                                                listCtx->variableIdentifier())
    {
        int lineNumber = idCtx->getStart()->getLine();
        string variableName = toLowerCase(idCtx->IDENTIFIER()->getText());
        SymtabEntry *variableId = symtabStack->lookupLocal(variableName);

        if (variableId == nullptr)
        {
            variableId = symtabStack->enterLocal(variableName, VARIABLE);
            variableId->setType(typeCtx->type);

            // Assign slot numbers to local variables.
            Symtab *symtab = variableId->getSymtab();
            if (symtab->getNestingLevel() > 1)
            {
                variableId->setSlotNumber(symtab->nextSlotNumber());
            }

            idCtx->entry = variableId;
        }
        else
        {
            error.flag(REDECLARED_IDENTIFIER, ctx);
        }

        variableId->appendLineNumber(lineNumber);
    }

    return nullptr;
}

Object Semantics::visitAssignmentStatement(
                                Pcl6Parser::AssignmentStatementContext *ctx)
{
    Pcl6Parser::LhsContext *lhsCtx = ctx->lhs();
    Pcl6Parser::RhsContext *rhsCtx = ctx->rhs();

    visitChildren(ctx);

    Typespec *lhsType = lhsCtx->type;
    Typespec *rhsType = rhsCtx->expression()->type;

    if (!TypeChecker::areAssignmentCompatible(lhsType, rhsType))
    {
        error.flag(INCOMPATIBLE_ASSIGNMENT, ctx);
    }

    return nullptr;
}

Object Semantics::visitLhs(Pcl6Parser::LhsContext *ctx)
{
    Pcl6Parser::VariableContext *varCtx = ctx->variable();
    visit(varCtx);
    ctx->type = varCtx->type;

    return nullptr;
}

Object Semantics::visitIfStatement(Pcl6Parser::IfStatementContext *ctx)
{
    Pcl6Parser::ExpressionContext     *exprCtx  = ctx->expression();
    Pcl6Parser::TrueStatementContext  *trueCtx  = ctx->trueStatement();
    Pcl6Parser::FalseStatementContext *falseCtx = ctx->falseStatement();

    visit(exprCtx);
    Typespec *expr_type = exprCtx->type;
    
    if (!TypeChecker::isBoolean(expr_type))
    {
        error.flag(TYPE_MUST_BE_BOOLEAN, exprCtx);
    }
    
    visit(trueCtx);
    if (falseCtx != nullptr) visit(falseCtx);

    return nullptr;
}

Object Semantics::visitCaseStatement(Pcl6Parser::CaseStatementContext *ctx)
{
    Pcl6Parser::ExpressionContext *exprCtx = ctx->expression();
    visit(exprCtx);
    Typespec *exprType = exprCtx->type;
    Form exprTypeForm = exprType->getForm();

    if (   (   (exprTypeForm != SCALAR)
            && (exprTypeForm != ENUMERATION)
            && (exprTypeForm != SUBRANGE))
        || (exprType == Predefined::realType)
        || (exprType == Predefined::stringType))
    {
        error.flag(TYPE_MISMATCH, exprCtx);
        exprType = Predefined::integerType;
    }

    set<int> constants;
    Pcl6Parser::CaseBranchListContext *branchListCtx = ctx->caseBranchList();

    // Loop over the CASE branches.
    for (Pcl6Parser::CaseBranchContext *branchCtx :
                                                branchListCtx->caseBranch())
    {
        Pcl6Parser::CaseConstantListContext *constListCtx =
                                                branchCtx->caseConstantList();
        Pcl6Parser::StatementContext *stmtCtx = branchCtx->statement();

        if (constListCtx != nullptr)
        {
            // Loop over the CASE constants in each branch.
            for (Pcl6Parser::CaseConstantContext *caseConstCtx :
                                                constListCtx->caseConstant())
            {
                Pcl6Parser::ConstantContext *constCtx =
                                                    caseConstCtx->constant();
                Object constValue = visit(constCtx);

                caseConstCtx->type  = constCtx->type;
                caseConstCtx->value = 0;

                if (constCtx->type != exprType)
                {
                    error.flag(TYPE_MISMATCH, constCtx);
                }
                else if (   (constCtx->type == Predefined::integerType)
                         || (constCtx->type->getForm() == ENUMERATION))
                {
                    caseConstCtx->value = constValue.as<int>();
                }
                else if (constCtx->type == Predefined::charType)
                {
                    caseConstCtx->value = constValue.as<char>();
                }

                if (constants.find(caseConstCtx->value) != constants.end())
                {
                    error.flag(DUPLICATE_CASE_CONSTANT, constCtx);
                }
                else
                {
                    constants.insert(caseConstCtx->value);
                }
            }
        }

        if (stmtCtx != nullptr) visit(stmtCtx);
    }

    return nullptr;
}

Object Semantics::visitRepeatStatement(
                                    Pcl6Parser::RepeatStatementContext *ctx)
{
    Pcl6Parser::ExpressionContext *exprCtx = ctx->expression();
    visit(exprCtx);
    Typespec *exprType = exprCtx->type;

    if (!TypeChecker::isBoolean(exprType))
    {
        error.flag(TYPE_MUST_BE_BOOLEAN, exprCtx);
    }

    visit(ctx->statementList());
    return nullptr;
}

Object Semantics::visitWhileStatement(Pcl6Parser::WhileStatementContext *ctx)
{
    Pcl6Parser::ExpressionContext *exprCtx = ctx->expression();
    visit(exprCtx);
    Typespec *exprType = exprCtx->type;

    if (!TypeChecker::isBoolean(exprType))
    {
        error.flag(TYPE_MUST_BE_BOOLEAN, exprCtx);
    }

    visit(ctx->statement());
    return nullptr;
}

Object Semantics::visitForStatement(Pcl6Parser::ForStatementContext *ctx)
{
    Pcl6Parser::VariableContext *varCtx = ctx->variable();
    visit(varCtx);

    string controlName = toLowerCase(varCtx->variableIdentifier()->getText());
    Typespec *controlType = Predefined::integerType;

    if (varCtx->entry != nullptr)
    {
        controlType = varCtx->type;
        if (   (controlType->getForm() != SCALAR )
            || (controlType == Predefined::realType)
            || (controlType == Predefined::stringType))
        {
            error.flag(INVALID_CONTROL_VARIABLE, varCtx);
        }
    }
    else
    {
        error.flag(UNDECLARED_IDENTIFIER, ctx->getStart()->getLine(),
                   controlName);
    }

    Pcl6Parser::ExpressionContext *startCtx = ctx->expression()[0];
    Pcl6Parser::ExpressionContext *endCtx   = ctx->expression()[1];

    visit(startCtx);
    visit(endCtx);

    if (startCtx->type != controlType)  error.flag(TYPE_MISMATCH, startCtx);
    if (startCtx->type != endCtx->type) error.flag(TYPE_MISMATCH, endCtx);

    visit(ctx->statement());
    return nullptr;
}

Object Semantics::visitExpression(Pcl6Parser::ExpressionContext *ctx)
{
    Pcl6Parser::SimpleExpressionContext *simpleCtx1 = ctx->simpleExpression()[0];

    // First simple expression.
    visit(simpleCtx1);

    Typespec *simpleType1 = simpleCtx1->type;
    ctx->type = simpleType1;

    Pcl6Parser::RelOpContext *relopCtx = ctx->relOp();

    // Second simple expression?
    if (relopCtx != nullptr)
    {
        Pcl6Parser::SimpleExpressionContext *simpleCtx2 =
                                                    ctx->simpleExpression()[1];
        visit(simpleCtx2);

        Typespec *simpleType2 = simpleCtx2->type;
        if (!TypeChecker::areComparisonCompatible(simpleType1, simpleType2))
        {
            error.flag(INCOMPATIBLE_COMPARISON, ctx);
        }

        ctx->type = Predefined::booleanType;
    }

    return nullptr;
}

Object Semantics::visitSimpleExpression(
                                    Pcl6Parser::SimpleExpressionContext *ctx)
{
    int count = ctx->term().size();
    Pcl6Parser::SignContext *signCtx = ctx->sign();
    bool hasSign = signCtx != nullptr;
    Pcl6Parser::TermContext *termCtx1 = ctx->term()[0];

    if (hasSign)
    {
        string sign = signCtx->getText();
        if ((sign== "+") && (sign == "-"))
        {
            error.flag(INVALID_SIGN, signCtx);
        }
    }

    // First term.
    visit(termCtx1);
    Typespec *termType1 = termCtx1->type;

    // Loop over any subsequent terms.
    for (int i = 1; i < count; i++)
    {
        string op = toLowerCase(ctx->addOp()[i-1]->getText());
        Pcl6Parser::TermContext *termCtx2 = ctx->term()[i];
        visit(termCtx2);
        Typespec *termType2 = termCtx2->type;

        // Both operands bool ==> bool result. Else type mismatch.
        if (op == "or")
        {
            if (!TypeChecker::isBoolean(termType1))
            {
                error.flag(TYPE_MUST_BE_BOOLEAN, termCtx1);
            }
            if (!TypeChecker::isBoolean(termType2))
            {
                error.flag(TYPE_MUST_BE_BOOLEAN, termCtx2);
            }
            if (hasSign)
            {
                error.flag(INVALID_SIGN, signCtx);
            }

            termType2 = Predefined::booleanType;
        }
        else if (op == "+")
        {
            // Both operands integer ==> integer result
            if (TypeChecker::areBothInteger(termType1, termType2))
            {
                termType2 = Predefined::integerType;
            }

            // Both real operands ==> real result
            // One real and one integer operand ==> real result
            else if (TypeChecker::isAtLeastOneReal(termType1, termType2))
            {
                termType2 = Predefined::realType;
            }

            // Both operands string ==> string result
            else if (TypeChecker::areBothString(termType1, termType2))
            {
                if (hasSign) error.flag(INVALID_SIGN, signCtx);
                termType2 = Predefined::stringType;
            }

            // Type mismatch.
            else
            {
                if (!TypeChecker::isIntegerOrReal(termType1))
                {
                    error.flag(TYPE_MUST_BE_NUMERIC, termCtx1);
                    termType2 = Predefined::integerType;
                }
                if (!TypeChecker::isIntegerOrReal(termType2))
                {
                    error.flag(TYPE_MUST_BE_NUMERIC, termCtx2);
                    termType2 = Predefined::integerType;
                }
            }
        }
        else  // -
        {
            // Both operands integer ==> integer result
            if (TypeChecker::areBothInteger(termType1, termType2))
            {
                termType2 = Predefined::integerType;
            }

            // Both real operands ==> real result
            // One real and one integer operand ==> real result
            else if (TypeChecker::isAtLeastOneReal(termType1, termType2))
            {
                termType2 = Predefined::realType;
            }

            // Type mismatch.
            else
            {
                if (!TypeChecker::isIntegerOrReal(termType1))
                {
                    error.flag(TYPE_MUST_BE_NUMERIC, termCtx1);
                    termType2 = Predefined::integerType;
                }
                if (!TypeChecker::isIntegerOrReal(termType2))
                {
                    error.flag(TYPE_MUST_BE_NUMERIC, termCtx2);
                    termType2 = Predefined::integerType;
                }
            }
        }

        termType1 = termType2;
    }

    ctx->type = termType1;
    return nullptr;
}

Object Semantics::visitTerm(Pcl6Parser::TermContext *ctx)
{
    int count = ctx->factor().size();
    Pcl6Parser::FactorContext *factorCtx1 = ctx->factor()[0];

    // First factor.
    visit(factorCtx1);
    Typespec *factorType1 = factorCtx1->type;

    // Loop over any subsequent factors.
    for (int i = 1; i < count; i++)
    {
        string op = toLowerCase(ctx->mulOp()[i-1]->getText());
        Pcl6Parser::FactorContext *factorCtx2 = ctx->factor()[i];
        visit(factorCtx2);
        Typespec *factorType2 = factorCtx2->type;

        if (op == "*")
        {
            // Both operands integer  ==> integer result
            if (TypeChecker::areBothInteger(factorType1, factorType2))
            {
                factorType2 = Predefined::integerType;
            }

            // Both real operands ==> real result
            // One real and one integer operand ==> real result
            else if (TypeChecker::isAtLeastOneReal(factorType1, factorType2))
            {
                factorType2 = Predefined::realType;
            }

            // Type mismatch.
            else
            {
                if (!TypeChecker::isIntegerOrReal(factorType1))
                {
                    error.flag(TYPE_MUST_BE_NUMERIC, factorCtx1);
                    factorType2 = Predefined::integerType;
                }
                if (!TypeChecker::isIntegerOrReal(factorType2))
                {
                    error.flag(TYPE_MUST_BE_NUMERIC, factorCtx2);
                    factorType2 = Predefined::integerType;
                }
            }
        }
        else if (op == "/")
        {
            // All integer and real operand combinations ==> real result
            if (   TypeChecker::areBothInteger(factorType1, factorType2)
                || TypeChecker::isAtLeastOneReal(factorType1, factorType2))
            {
                factorType2 = Predefined::realType;
            }

            // Type mismatch.
            else
            {
                if (!TypeChecker::isIntegerOrReal(factorType1))
                {
                    error.flag(TYPE_MUST_BE_NUMERIC, factorCtx1);
                    factorType2 = Predefined::integerType;
                }
                if (!TypeChecker::isIntegerOrReal(factorType2))
                {
                    error.flag(TYPE_MUST_BE_NUMERIC, factorCtx2);
                    factorType2 = Predefined::integerType;
                }
            }
        }
        else if ((op == "div") ||(op == "mod"))
        {
            // Both operands integer ==> integer result. Else type mismatch.
            if (!TypeChecker::isInteger(factorType1))
            {
                error.flag(TYPE_MUST_BE_INTEGER, factorCtx1);
                factorType2 = Predefined::integerType;
            }
            if (!TypeChecker::isInteger(factorType2))
            {
                error.flag(TYPE_MUST_BE_INTEGER, factorCtx2);
                factorType2 = Predefined::integerType;
            }

            ctx->type = Predefined::integerType;
        }
        else if (op == "and")
        {
            // Both operands bool ==> bool result. Else type mismatch.
            if (!TypeChecker::isBoolean(factorType1))
            {
                error.flag(TYPE_MUST_BE_BOOLEAN, factorCtx1);
                factorType2 = Predefined::booleanType;
            }
            if (!TypeChecker::isBoolean(factorType2))
            {
                error.flag(TYPE_MUST_BE_BOOLEAN, factorCtx2);
                factorType2 = Predefined::booleanType;
            }
        }

        factorType1 = factorType2;
    }

    ctx->type = factorType1;
    return nullptr;
}

Object Semantics::visitVariableFactor(Pcl6Parser::VariableFactorContext *ctx)
{
    Pcl6Parser::VariableContext *varCtx = ctx->variable();
    visit(varCtx);
    ctx->type = varCtx->type;

    return nullptr;
}

Object Semantics::visitVariable(Pcl6Parser::VariableContext *ctx)
{
    Pcl6Parser::VariableIdentifierContext *varIdCtx =
                                                    ctx->variableIdentifier();

    visit(varIdCtx);
    ctx->entry = varIdCtx->entry;
    ctx->type  = variableDatatype(ctx, varIdCtx->type);

    return nullptr;
}

Object Semantics::visitVariableIdentifier(
                                Pcl6Parser::VariableIdentifierContext *ctx)
{

    string variableName = toLowerCase(ctx->IDENTIFIER()->getText());
    SymtabEntry *variableId = symtabStack->lookup(variableName);

    if (variableId != nullptr)
    {
        int lineNumber = ctx->getStart()->getLine();
        ctx->type = variableId->getType();
        ctx->entry = variableId;
        variableId->appendLineNumber(lineNumber);

        Kind kind = variableId->getKind();
        switch (kind)
        {
            case TYPE:
            case PROGRAM:
            case PROGRAM_PARAMETER:
            case PROCEDURE:
            case UNDEFINED:
                error.flag(INVALID_VARIABLE, ctx);
                break;

            default: break;
        }
    }
    else
    {
        error.flag(UNDECLARED_IDENTIFIER, ctx);
        ctx->type = Predefined::integerType;
    }

    return nullptr;
}

Typespec *Semantics::variableDatatype(Pcl6Parser::VariableContext *varCtx,
                                      Typespec *varType)
{
    Typespec *type = varType;

    for (Pcl6Parser::ModifierContext *modCtx : varCtx->modifier())
    {
        // Subscripts.
        if (modCtx->indexList() != nullptr)
        {
            Pcl6Parser::IndexListContext *indexListCtx = modCtx->indexList();

            // Loop over the subscripts.
            for (Pcl6Parser::IndexContext *indexCtx : indexListCtx->index())
            {
                if (type->getForm() == ARRAY)
                {
                    Typespec *indexType = type->getArrayIndexType();
                    Pcl6Parser::ExpressionContext *exprCtx =
                                                        indexCtx->expression();
                    visit(exprCtx);

                    if (indexType->baseType() != exprCtx->type->baseType())
                    {
                        error.flag(TYPE_MISMATCH, exprCtx);
                    }

                    // Datatype of the next dimension.
                    type = type->getArrayElementType();
                }
                else
                {
                    error.flag(TOO_MANY_SUBSCRIPTS, indexCtx);
                }
            }
        }
        else // Record field.
        {
            if (type->getForm() == RECORD)
            {
                Symtab *symtab = type->getRecordSymtab();
                Pcl6Parser::FieldContext *fieldCtx = modCtx->field();
                string fieldName =
                                toLowerCase(fieldCtx->IDENTIFIER()->getText());
                SymtabEntry *fieldId = symtab->lookup(fieldName);

                if (fieldId != nullptr)
                {
                    type = fieldId->getType();
                    fieldCtx->entry = fieldId;
                    fieldCtx->type = type;
                    fieldId->appendLineNumber(modCtx->getStart()->getLine());
                }
                else
                {
                    error.flag(INVALID_FIELD, modCtx);
                }
            }

            // Not a record variable.
            else
            {
                error.flag(INVALID_FIELD, modCtx);
            }
        }
    }

    return type;
}

Object Semantics::visitNumberFactor(Pcl6Parser::NumberFactorContext *ctx)
{
    Pcl6Parser::NumberContext          *numberCtx   = ctx->number();
    Pcl6Parser::UnsignedNumberContext  *unsignedCtx = numberCtx->unsignedNumber();
    Pcl6Parser::IntegerConstantContext *integerCtx  = unsignedCtx->integerConstant();

    ctx->type = (integerCtx != nullptr) ? Predefined::integerType
                                        : Predefined::realType;

    return nullptr;
}

Object Semantics::visitCharacterFactor(Pcl6Parser::CharacterFactorContext *ctx)
{
    ctx->type = Predefined::charType;
    return nullptr;
}

Object Semantics::visitStringFactor(Pcl6Parser::StringFactorContext *ctx)
{
    ctx->type = Predefined::stringType;
    return nullptr;
}

Object Semantics::visitNotFactor(Pcl6Parser::NotFactorContext *ctx)
{
    Pcl6Parser::FactorContext *factorCtx = ctx->factor();
    visit(factorCtx);

    if (factorCtx->type != Predefined::booleanType)
    {
        error.flag(TYPE_MUST_BE_BOOLEAN, factorCtx);
    }

    ctx->type = Predefined::booleanType;
    return nullptr;
}

Object Semantics::visitParenthesizedFactor(
                                Pcl6Parser::ParenthesizedFactorContext *ctx)
{
    Pcl6Parser::ExpressionContext *exprCtx = ctx->expression();
    visit(exprCtx);
    ctx->type = exprCtx->type;

    return nullptr;
}

} // namespace frontend
